from __future__ import annotations

from collections.abc import Generator, Iterable, MutableSet
from typing import Any, TypeVar

from jax.tree_util import register_pytree_with_keys
from rich.tree import Tree

from .annotations import PyTree
from .display.display_generic import _verify, display_class, display_generic

__all__: list[str] = []


def graph_arrow(directed: bool) -> str:  # noqa: FBT001
    return '⟶' if directed else '↔'


def graph_edge_name(arrow: str, source: str, target: str) -> str:
    return f"{source}{arrow}{target}"


try:
    import networkx as nx
except ImportError:
    msg = "NetworkX not available"
    def register_graph_as_jax_pytree(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
    def register_graph_as_nnx_node(graph_type: type[Any]) -> None:
        raise RuntimeError(msg)
else:
    T = TypeVar('T', bound="nx.Graph[Any]")

    def validate_node_name(name: Any,
                           arrow: str
                           ) -> None:
        if not isinstance(name, str):
            raise TypeError
        if arrow in name:
            raise ValueError

    def flatten_helper(graph: nx.Graph[Any],
                       arrow: str
                       ) -> Generator[tuple[str, Any], None, None]:
        for name, data in sorted(graph.nodes.data()):
            validate_node_name(name, arrow)
            yield name, data

        def undirected_edge_key(source_target_data: tuple[str, str, Any]) -> tuple[str, ...]:
            source, target, _ = source_target_data
            return tuple(sorted([source, target]))
        edge_data = graph.edges.data()
        directed = isinstance(graph, nx.DiGraph)
        if not directed:
            edge_data = sorted(edge_data, key=undirected_edge_key)
        for source, target, data in edge_data:
            validate_node_name(source, arrow)
            validate_node_name(target, arrow)
            new_source, new_target = ((source, target)
                                      if directed
                                      else sorted([source, target]))
            yield graph_edge_name(arrow, new_source, new_target), data

    def init_graph(graph: nx.Graph[Any],
                   items: Iterable[tuple[str, Any]],
                   arrow: str
                   ) -> None:
        for key, value in items:
            if not isinstance(value, dict):
                raise TypeError
            if arrow in key:
                source, target = key.split(arrow, 2)
                graph.add_edge(source, target, **value)
            else:
                graph.add_node(key, **value)

    def register_graph_as_jax_pytree(graph_type: type[nx.Graph[Any]]) -> None:  # pyright: ignore
        arrow = graph_arrow(issubclass(graph_type, nx.DiGraph))

        def unflatten_tree(names: tuple[str, ...], values: Iterable[Any], /) -> nx.Graph[Any]:
            graph = graph_type()
            init_graph(graph, zip(names, values, strict=True), arrow)
            return graph

        def flatten_with_keys(graph: nx.Graph[Any], /
                              ) -> tuple[Iterable[tuple[str, Any]], tuple[str, ...]]:
            names_and_values = tuple(flatten_helper(graph, arrow))
            names = tuple(name for name, _ in names_and_values)
            return (names_and_values, names)

        def flatten_tree(graph: nx.Graph[Any], /) -> tuple[Iterable[PyTree], tuple[str, ...]]:
            names_and_values = tuple(flatten_helper(graph, arrow))
            names, values = zip(*names_and_values, strict=True)
            return values, names

        register_pytree_with_keys(graph_type, flatten_with_keys, unflatten_tree, flatten_tree)

    @display_generic.register(nx.Graph)
    def _(value: nx.Graph[Any],
          *,
          seen: MutableSet[int] | None = None,
          key: str = '',
          ) -> Tree:
        if seen is None:
            seen = set()
        with _verify(value, seen, key) as x:
            if x:
                return x
            arrow = graph_arrow(isinstance(value, nx.DiGraph))
            retval = display_class(key, type(value))
            for name, node in value.nodes.items():
                retval.children.append(display_generic(node, seen=seen, key=name))
            for (source, target), edge in value.edges.items():
                key = graph_edge_name(arrow, source, target)
                retval.children.append(display_generic(edge, seen=seen, key=key))
            return retval

    try:
        from flax.nnx.nnx.graph import register_graph_node_type
        from flax.typing import Key
    except ImportError:
        msg = "NNX not available"
        def register_graph_as_nnx_node(graph_type: type[Any]) -> None:
            raise RuntimeError(msg)
    else:
        def register_graph_as_nnx_node(graph_type: type[T]) -> None:  # pyright: ignore
            arrow = graph_arrow(issubclass(graph_type, nx.DiGraph))

            # flatten: Callable[[Node], tuple[Sequence[tuple[str, Leaf]], AuxData]],
            def flatten_graph(graph: T, /) -> tuple[tuple[tuple[str, Any], ...], None]:
                t = tuple(flatten_helper(graph, arrow))
                return t, None

            # set_key: Callable[[Node, Key, Leaf], None],
            def set_key_graph(graph: T, key: Key, value: Any, /) -> None:
                if not isinstance(value, dict):
                    raise TypeError
                assert isinstance(key, str)
                d: dict[str, Any]
                if arrow in key:
                    source, target = key.split(arrow, 2)
                    if not graph.has_edge(source, target):
                        graph.add_edge(source, target, **value)
                        return
                    d = graph.edges[source, target]
                elif not graph.has_node(key):
                    graph.add_node(key, **value)
                    return
                else:
                    d = graph.nodes[key]
                d.clear()
                d.update(value)

            # pop_key: Callable[[Node, Key], Leaf],
            def pop_key_graph(graph: T, key: Key, /) -> Any:
                assert isinstance(key, str)
                if arrow in key:
                    source, target = key.split(arrow, 2)
                    retval = graph.edges[source, target]
                    graph.remove_edge(source, target)
                    return retval
                retval = graph.nodes[key]
                graph.remove_node(key)
                return retval

            # create_empty: Callable[[AuxData], Node],
            def create_empty_graph(metadata: None, /) -> T:
                del metadata
                return graph_type()

            # clear: Callable[[Node, AuxData], None],
            def clear_graph(graph: T, /) -> None:
                graph.clear()

            register_graph_node_type(graph_type,
                                     flatten_graph,
                                     set_key_graph,
                                     pop_key_graph,
                                     create_empty_graph,
                                     clear_graph)



from __future__ import annotations

from functools import partial, singledispatch
from math import prod
from numbers import Complex, Integral, Real
from typing import Any

import jax
import jax.numpy as jnp
import numpy as np
from jax import tree

from .annotations import Array, PyTree
from .dtypes import default_tols


def assert_tree_allclose(actual: PyTree,
                         desired: PyTree,
                         original_name: str | None = None,
                         original_value: PyTree | None = None,
                         *,
                         rtol: float | None = None,
                         atol: float | None = None) -> None:
    """Assert that an actual pytree matches a desired pytree.

    If the assertion fails, a passing test string is printed::

    ```python
    from tjax import assert_tree_allclose, RealNumeric
    from tjax.dataclasses import dataclass

    @dataclass
    class A:
        x: RealNumeric
        y: RealNumeric

    @dataclass
    class B:
        z: A

    original = B(A(1.2, 3.4))
    desired = B(A(3.0, 4.0))
    actual = B(A(1.2, 5.2))

    assert_tree_allclose(actual, desired)
    ```
    This prints::
    ```
    Tree leaves don't match at position 0 with rtol=0.0001 and atol=1e-06.
    Mismatched elements: 1 / 1 (100%)
    Maximum absolute difference: 1.8
    Maximum relative difference: 0.6

    Actual: B(z=A(x=1.2, y=5.2))
    Desired: B(z=A(x=3.0, y=4.0))
    Test string:
    desired = B(A(1.2, 5.2))
    ```
    The test string can then be pasted.

    Args:
        actual: The actual value.
        desired: The desired value.
        original_name: The variable name that contains the original value.
        original_value: The original value.  This is usually a pytree like a dataclass that has the
            same type as actual and desired, but contains different values.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    """
    flattened_actual, structure_actual = tree.flatten(actual)
    flattened_desired, structure_desired = tree.flatten(desired)
    if structure_actual != structure_desired:
        msg = f"\nTree structure mismatch.\nActual: {actual}\nDesired: {desired}\n"
        raise AssertionError(msg)

    for i, (actual_, desired_) in enumerate(zip(flattened_actual, flattened_desired, strict=True)):
        dtype = jnp.result_type(actual_, desired_)
        tols = default_tols(dtype.type, rtol=rtol, atol=atol)
        try:
            np.testing.assert_allclose(actual_, desired_, **tols)
        except AssertionError as exception:
            old_message = exception.args[0].split('\n')
            best_part_of_old_message = "\n".join(old_message[3:6]).replace("Max ", "Maximum ")
            test_string = (get_relative_test_string(actual, original_name, original_value, **tols)
                           if original_name is not None and original_value is not None
                           else get_test_string(actual, **tols))
            test_string = "desired = " + test_string
            # style_config = yapf.style.CreatePEP8Style()
            # style_config['COLUMN_LIMIT'] = column_limit
            # test_string = yapf.yapf_api.FormatCode(test_string, style_config=style_config)[0]
            message = (
                f"\nTree leaves don't match at position {i} with rtol={tols['rtol']} and "
                f"atol={tols['atol']}.\n"
                f"{best_part_of_old_message}\n\n"
                f"Actual: {actual}\nDesired: {desired}\n"
                f"Test string:\n{test_string}")
            raise AssertionError(message) from None


def tree_allclose(actual: PyTree,
                  desired: PyTree,
                  rtol: float | None = None,
                  atol: float | None = None) -> bool:
    """Return whether two pytrees are close.

    Args:
        actual: The actual value.
        desired: The desired value.
        rtol: The relative tolerance of the comparisons in the comparison.
        atol: The absolute tolerance of the comparisons in the comparison.
    """
    def allclose(actual_array: Array, desired_array: Array) -> bool:
        dtype = jnp.result_type(actual_array, desired_array)
        tols = default_tols(dtype.type, rtol=rtol, atol=atol)
        return bool(jnp.allclose(actual_array, desired_array, **tols))

    return tree.all(tree.map(allclose, actual, desired))


# get test string ----------------------------------------------------------------------------------
# Redefinition typing errors in this file are due to https://github.com/python/mypy/issues/2904.
@singledispatch
def get_test_string(actual: Any, rtol: float, atol: float) -> str:
    """Produce a short string of Python code that produces the actual value.

    Args:
        actual: The actual value that was produced.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.
    """
    return repr(actual)


@get_test_string.register(np.ndarray)
@get_test_string.register(jax.Array)
def _(actual: Array | jax.Array, rtol: float, atol: float) -> str:
    if prod(actual.shape) == 0:
        return f"np.empty({actual.shape}, dtype=np.{actual.dtype})"
    with np.printoptions(formatter={'float_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                          atol=atol),
                                    'complex_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                            atol=atol)}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace(
            '  ', ' ').replace('dtype=', 'dtype=np.')


@get_test_string.register
def _(actual: Complex, rtol: float, atol: float) -> str:
    x = float(actual) if isinstance(actual, Real) else complex(actual)
    return _inexact_number_to_string(x, rtol, atol)


@get_test_string.register
def _(actual: Integral, rtol: float, atol: float) -> str:
    return str(actual)


@get_test_string.register(list)
@get_test_string.register(tuple)
def _(actual: list[Any] | tuple[Any], rtol: float, atol: float) -> str:
    is_list = isinstance(actual, list)
    is_named_tuple = not is_list and type(actual).__name__ != 'tuple'
    return ((type(actual).__name__ if is_named_tuple else "")
            + ("[" if is_list else "(")
            + ", ".join(get_test_string(sub_actual, rtol, atol)
                        for i, sub_actual in enumerate(actual))
            + (',' if len(actual) == 1 else '')
            + ("]" if is_list else ")"))


@get_test_string.register(dict)
def _(actual: dict[Any, Any], rtol: float, atol: float) -> str:
    return '{' + ",\n".join(repr(key) + ': ' + get_test_string(sub_actual, rtol, atol)
                            for key, sub_actual in actual.items()) + '}'


# get relative test string -------------------------------------------------------------------------
@singledispatch
def get_relative_test_string(actual: Any,
                             original_name: str,
                             original: Any,
                             rtol: float,
                             atol: float) -> str:
    """Produce code for use in tests based on actual and produced values.

    Args:
        original_name: The name of the variable containing an original value.
        actual: The actual value that was produced, and that should be the desired value.
        original: The original value.
        rtol: The relative tolerance of the comparisons in the assertion.
        atol: The absolute tolerance of the comparisons in the assertion.

    Returns:
        A string of Python code that produces the desired value from an "original value" (could be
        zeroed-out, for example).
    """
    return str(actual)


@get_relative_test_string.register(np.ndarray)
@get_relative_test_string.register(jax.Array)
def _(actual: Array | jax.Array, original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    with np.printoptions(formatter={'float_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                          atol=atol),
                                    'complex_kind': partial(_inexact_number_to_string, rtol=rtol,
                                                            atol=atol)}):
        return "np." + repr(np.asarray(actual)).replace(' ]', ']').replace(' ,', ',').replace('  ',
                                                                                              ' ')


@get_relative_test_string.register
def _(actual: Complex, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return _inexact_number_to_string(actual, rtol, atol)  # type: ignore[arg-type] # pyright: ignore


@get_relative_test_string.register
def _(actual: Integral, original_name: str, original: Any, rtol: float, atol: float) -> str:
    return str(actual)


@get_relative_test_string.register(list)
@get_relative_test_string.register(tuple)
def _(actual: list[Any] | tuple[Any], original_name: str, original: Any, rtol: float,
      atol: float) -> str:
    is_list = isinstance(actual, list)
    return (("[" if is_list else "(")
            + ", ".join(get_relative_test_string(f"{original_name}[{i}]",
                                                 sub_actual, sub_original,
                                                 rtol, atol)
                        for i, (sub_actual, sub_original) in enumerate(zip(actual, original,
                                                                           strict=True)))
            + ("]" if is_list else ")"))


@get_relative_test_string.register(dict)
def _(actual: dict[Any, Any], original_name: str, original: Any, rtol: float, atol: float) -> str:
    if not isinstance(original, dict):
        raise TypeError

    def relative_string(key: Any, sub_actual: Any) -> str:
        return get_relative_test_string(
            f"{original_name}[{key}]", sub_actual, original[key], rtol, atol)

    return '{' + ",\n".join(repr(key) + ': ' + relative_string(key, sub_actual)
                            for key, sub_actual in actual.items()) + '}'


# Private functions --------------------------------------------------------------------------------
def _float_to_string_with_precision(x: complex, precision: int) -> str:
    with np.printoptions(precision=precision, floatmode='maxprec'):
        return repr(np.asarray(x))[6:-1]


def _inexact_number_to_string(x: complex | np.inexact[Any], rtol: float, atol: float) -> str:
    y: float | complex
    if isinstance(x, Real):  # type: ignore[unreachable]
        y = float(x)  # type: ignore[unreachable]
    elif isinstance(x, Complex):
        y = complex(x)
    else:
        raise TypeError
    retval = ""
    for i in range(20):
        retval = _float_to_string_with_precision(y, i)
        if np.isclose(complex(retval), x, rtol=rtol, atol=atol):
            break
    return retval



from __future__ import annotations

from typing import Any

import jax
import jax.numpy as jnp

from .annotations import JaxBooleanArray


def dynamic_tree_all(tree: Any) -> JaxBooleanArray:
    """Like `jax.tree.all`, but can be used in dynamic code like jitted functions and loops."""
    return jax.tree.reduce(jnp.logical_and, tree, jnp.asarray(True))



from __future__ import annotations

from typing import Any

import numpy as np
from jax import debug, tree
from rich.console import Console

from .internal import internal_print_generic


def print_generic(*args: Any,
                  raise_on_nan: bool = True,
                  immediate: bool = False,
                  console: Console | None = None,
                  **kwargs: Any
                  ) -> None:
    """Uses internal_print_generic in a tapped function.

    Args:
        raise_on_nan: Assert if NaN is found in the output.
        immediate: Print the value immediately even if they're tracers.  Do not wait for the
            compiled function to be called.
        console: The console that formats the output.
        args: Positional arguments to be printed.  Only dynamic arguments are allowed.
        kwargs: Keyword arguments to be printed.  Only static keys and dynamic values are allowed.
    """
    if immediate:
        internal_print_generic(*args, raise_on_nan=raise_on_nan, console=console, **kwargs)
        return

    leaves, tree_def = tree.flatten((args, kwargs))

    def fix(u_arg: Any, arg: Any) -> Any:
        return (np.asarray(u_arg) if isinstance(arg, np.ndarray)
                else int(u_arg.item()) if isinstance(arg, int)
                else u_arg)

    def callback(*callback_leaves: Any) -> None:
        unflattened_tree = tree.unflatten(tree_def, callback_leaves)
        v_args, v_kwargs = tree.map(fix, unflattened_tree, (args, kwargs))
        internal_print_generic(*v_args, raise_on_nan=raise_on_nan, console=console, **v_kwargs)

    debug.callback(callback, *leaves, ordered=True)



from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypeVar

from typing_extensions import ParamSpec, override

from .function_markers import abstract_custom_jvp_marker, abstract_jit_marker
from .shims import custom_jvp_method, jit

R_co = TypeVar('R_co', covariant=True)
F = TypeVar('F', bound=Callable[..., Any])
P = ParamSpec('P')
U = TypeVar("U")


class JaxAbstractClass:
    """A class with abstract methods whose overrides need to be decorated with Jax decorators."""
    @override
    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        super_cls = super(cls, cls)
        for name, original_method in list(cls.__dict__.items()):
            if not hasattr(super_cls, name):
                continue
            super_method = getattr(super_cls, name)
            if super_method is original_method:
                continue
            # Override detected.
            method = original_method
            if getattr(super_method, abstract_jit_marker, False):
                method = jit(original_method)
            if hasattr(super_method, abstract_custom_jvp_marker):
                jvp, nondiff_argnums = super_method._abstract_custom_jvp  # noqa: SLF001
                method_jvp: Any = custom_jvp_method(method, nondiff_argnums=nondiff_argnums)
                method_jvp.defjvp(jvp)
                setattr(cls, f'_original_{name}', method)
                method = method_jvp
            setattr(cls, name, method)


def abstract_jit(fun: F) -> F:
    """An abstract method whose override need to be jitted."""
    fun._abstract_jit = True  # type: ignore # noqa: SLF001 # pyright: ignore
    return fun


def abstract_custom_jvp(jvp: Callable[..., tuple[R_co, R_co]],
                        nondiff_argnums: tuple[int, ...] = ()
    ) -> Callable[[Callable[P, R_co]], Callable[P, R_co]]:
    """An abstract method whose override need to be decorated with custom_jvp_method."""
    def decorator(fun: Callable[P, R_co]) -> Callable[P, R_co]:
        fun._abstract_custom_jvp = (jvp,  # type: ignore # noqa: SLF001 # pyright: ignore
                                    nondiff_argnums)
        return fun
    return decorator



from __future__ import annotations

from typing import Any, TypeAlias, TypedDict

import jax.numpy as jnp


class Tols(TypedDict):
    rtol: float
    atol: float


# Work around jax.numpy._ScalarMeta
JNumber: TypeAlias = type[jnp.number[Any]]
rtol_dict: dict[JNumber, float] = {jnp.bfloat16.dtype.type: 1e-4, jnp.float16.dtype.type: 1e-3,
                                   jnp.float32.dtype.type: 1e-4, jnp.float64.dtype.type: 1e-5,
                                   jnp.complex64.dtype.type: 1e-4, jnp.complex128.dtype.type: 1e-5}
atol_dict: dict[JNumber, float] = {jnp.bfloat16.dtype.type: 1e-2, jnp.float16.dtype.type: 1e-3,
                                   jnp.float32.dtype.type: 1e-6, jnp.float64.dtype.type: 1e-8,
                                   jnp.complex64.dtype.type: 1e-6, jnp.complex128.dtype.type: 1e-8}


def default_rtol(dtype: JNumber) -> float:
    if not jnp.issubdtype(dtype, jnp.inexact):
        return 0.0
    return rtol_dict[dtype]


def default_atol(dtype: JNumber) -> float:
    if not jnp.issubdtype(dtype, jnp.inexact):
        return 1.0
    return atol_dict[dtype]


def default_tols(dtype: type[jnp.number[Any]],
                 *,
                 rtol: float | None = None,
                 atol: float | None = None) -> Tols:
    return Tols(rtol=default_rtol(dtype) if rtol is None else rtol,
                atol=default_atol(dtype) if atol is None else atol)



from __future__ import annotations

from collections.abc import Callable
from typing import overload

import jax.numpy as jnp
import numpy as np
from array_api_compat import get_namespace
from jax import Array
from jax.dtypes import canonicalize_dtype
from jax.lax import scan
from jax.random import normal

from .annotations import (ComplexArray, IntegralArray, JaxComplexArray, JaxIntegralArray,
                          JaxRealArray, KeyArray, RealArray)
from .dataclasses.dataclass import dataclass


@overload
def leaky_integrate(value: JaxComplexArray,
                    time_step: JaxRealArray,
                    drift: JaxComplexArray | None = None,
                    decay: JaxComplexArray | None = None,
                    *,
                    leaky_average: bool = False
                    ) -> JaxComplexArray:
    ...


@overload
def leaky_integrate(value: ComplexArray,
                    time_step: RealArray,
                    drift: ComplexArray | None = None,
                    decay: ComplexArray | None = None,
                    *,
                    leaky_average: bool = False
                    ) -> ComplexArray:
    ...


def leaky_integrate(value: ComplexArray,
                    time_step: RealArray,
                    drift: ComplexArray | None = None,
                    decay: ComplexArray | None = None,
                    *,
                    leaky_average: bool = False
                    ) -> ComplexArray:
    """Update the value so that it is the leaky integral (or leaky average).

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        decay: If provided, must have positive real component, and the value decays by exp(-decay)
            every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    if drift is None:
        if decay is None:
            return value
        xp = get_namespace(decay, time_step, value)
        return xp.exp(-decay * time_step) * value

    if decay is None:
        if leaky_average:
            raise ValueError
        xp = get_namespace(drift, time_step, value)
        return xp.asarray(value + drift * time_step)

    xp = get_namespace(drift, decay, time_step, value)
    scaled_integrand = (drift / decay) * -xp.expm1(-decay * time_step)

    if leaky_average:
        scaled_integrand *= decay.real

    return xp.exp(-decay * time_step) * value + scaled_integrand


@overload
def leaky_data_weight(iterations_times_time_step: JaxRealArray | JaxIntegralArray,
                      decay: JaxRealArray
                      ) -> JaxRealArray:
    ...


@overload
def leaky_data_weight(iterations_times_time_step: RealArray | IntegralArray,
                      decay: RealArray
                      ) -> RealArray:
    ...


def leaky_data_weight(iterations_times_time_step: RealArray | IntegralArray,
                      decay: RealArray
                      ) -> RealArray:
    """The amount of data that has been incorporated and has not been decayed.

    This equals leaky_integrate(0.0, iterations_times_time_step, 1.0, decay, leaky_average=True).
    """
    xp = get_namespace(iterations_times_time_step, decay)
    return -xp.expm1(-iterations_times_time_step * decay)


def diffused_leaky_integrate(value: JaxComplexArray,
                             time_step: JaxRealArray,
                             rng: KeyArray,
                             diffusion: JaxRealArray,
                             drift: JaxComplexArray | None = None,
                             decay: JaxComplexArray | None = None,
                             *,
                             leaky_average: bool = False
                             ) -> ComplexArray:
    """Update an Ornstein-Uhlenbeck process.

    Args:
        value: The current value of the leaky integral or average.
        time_step: The number of seconds that have passed.
        rng: The key array for the stochastic process.
        diffusion: The diffusion for the stochastic process.
        decay: If provided, the value decays by exp(-decay) every second.
        drift: If provided, the value increases by this every second.
        leaky_average: A flag indicating a leaky average rather than a leaky integral.  This scales
            the drift by the real component (in case the decay is complex) of the decay.
    """
    variance = (diffusion * time_step
                if decay is None
                else diffusion / decay * -jnp.expm1(-decay * time_step))
    jump = jnp.sqrt(variance) * normal(rng, value.shape)
    return leaky_integrate(value, time_step, drift, decay, leaky_average=leaky_average) + jump


@dataclass
class _FilterCarry:
    iterations: JaxRealArray
    value: JaxComplexArray


@overload
def leaky_integrate_time_series(time_series: IntegralArray | RealArray,
                                decay: RealArray
                                ) -> RealArray: ...
@overload
def leaky_integrate_time_series(time_series: IntegralArray | ComplexArray,
                                decay: ComplexArray
                                ) -> ComplexArray: ...
def leaky_integrate_time_series(time_series: IntegralArray | ComplexArray,
                                decay: ComplexArray
                                ) -> ComplexArray:
    xp = get_namespace(time_series, decay)
    dtype = jnp.promote_types(canonicalize_dtype(time_series.dtype), jnp.float32)
    time_series = jnp.asarray(time_series, dtype=dtype)
    decay = jnp.asarray(decay)
    if issubclass(time_series.dtype.type, np.integer):
        msg = "Cast the time series to a floating type."
        raise TypeError(msg)

    def g(carry: _FilterCarry, drift: JaxComplexArray) -> tuple[_FilterCarry, JaxComplexArray]:
        new_iterations = carry.iterations + 1.0
        data_weight = leaky_data_weight(new_iterations, decay.real)
        new_value = leaky_integrate(carry.value, jnp.asarray(1.0), drift, decay, leaky_average=True)
        assert isinstance(new_value, Array)
        new_carry = _FilterCarry(new_iterations, new_value)
        outputted_value = new_value / data_weight
        return new_carry, outputted_value

    # Cast the dtype from integer to floating point to prevent integer rounding.
    initial_value = jnp.zeros(time_series[0].shape, dtype=time_series.dtype)
    initial_carry = _FilterCarry(jnp.asarray(0.0), initial_value)

    _, filtered_time_series = scan(g, initial_carry, time_series)
    return xp.asarray(filtered_time_series)


def leaky_covariance(x_time_series: ComplexArray,
                     y_time_series: ComplexArray,
                     decay: ComplexArray,
                     *,
                     covariance_matrix: bool = False) -> ComplexArray:
    get_namespace(x_time_series, y_time_series, decay)  # Verify namespace.
    times: Callable[[ComplexArray, ComplexArray], ComplexArray]
    if covariance_matrix:
        if x_time_series.shape[0] != y_time_series.shape[0]:
            raise ValueError
        s = (np.newaxis,)

        def times(a: ComplexArray, b: ComplexArray, /) -> ComplexArray:
            a_selected = a[..., *(s * (b.ndim - 1))]
            b_selected = b[:, *(s * (a.ndim - 1))]
            return a_selected * b_selected  # type: ignore # pyright: ignore
    else:
        if x_time_series.shape != y_time_series.shape:
            raise ValueError

        def times(a: ComplexArray, b: ComplexArray, /) -> ComplexArray:
            return a * b  # type: ignore # pyright: ignore
    x = leaky_integrate_time_series(x_time_series, decay)
    y = leaky_integrate_time_series(y_time_series, decay)
    xy = leaky_integrate_time_series(times(x_time_series, y_time_series), decay)
    return xy - times(x, y)  # type: ignore # pyright: ignore



"""Tools for the differential programming library [JAX](https://github.com/google/jax)."""
from . import dataclasses, gradient
from ._src.abstract_method_decorators import JaxAbstractClass, abstract_custom_jvp, abstract_jit
from ._src.annotations import (Array, BooleanArray, BooleanNumeric, Complex, ComplexArray,
                               ComplexNumeric, Integral, IntegralArray, IntegralNumeric, JaxArray,
                               JaxBooleanArray, JaxComplexArray, JaxIntegralArray, JaxRealArray,
                               KeyArray, NumpyArray, NumpyBooleanArray, NumpyBooleanNumeric,
                               NumpyComplexArray, NumpyComplexNumeric, NumpyIntegralArray,
                               NumpyIntegralNumeric, NumpyRealArray, NumpyRealNumeric, PyTree, Real,
                               RealArray, RealNumeric, Shape, ShapeLike, SliceLike,
                               TapFunctionTransforms)
from ._src.cotangent_tools import (copy_cotangent, cotangent_combinator, print_cotangent,
                                   replace_cotangent, scale_cotangent, zero_tangent_like)
from ._src.display.display_generic import display_generic
from ._src.display.internal import internal_print_generic
from ._src.display.print_generic import print_generic
from ._src.dtypes import default_atol, default_rtol, default_tols
from ._src.graph import (graph_arrow, graph_edge_name, register_graph_as_jax_pytree,
                         register_graph_as_nnx_node)
from ._src.leaky_integral import (diffused_leaky_integrate, leaky_covariance, leaky_data_weight,
                                  leaky_integrate, leaky_integrate_time_series)
from ._src.math_tools import (abs_square, create_diagonal_array, divide_nonnegative, divide_where,
                              inverse_softplus, matrix_dot_product, matrix_vector_mul,
                              outer_product, softplus)
from ._src.partial import Partial
from ._src.shims import custom_jvp, custom_jvp_method, custom_vjp, custom_vjp_method, hessian, jit
from ._src.testing import (assert_tree_allclose, get_relative_test_string, get_test_string,
                           tree_allclose)

__all__ = [
    'Array',
    'BooleanArray',
    'BooleanNumeric',
    'Complex',
    'ComplexArray',
    'ComplexNumeric',
    'Integral',
    'IntegralArray',
    'IntegralNumeric',
    'JaxAbstractClass',
    'JaxArray',
    'JaxBooleanArray',
    'JaxComplexArray',
    'JaxIntegralArray',
    'JaxRealArray',
    'KeyArray',
    'NumpyArray',
    'NumpyBooleanArray',
    'NumpyBooleanNumeric',
    'NumpyComplexArray',
    'NumpyComplexNumeric',
    'NumpyIntegralArray',
    'NumpyIntegralNumeric',
    'NumpyRealArray',
    'NumpyRealNumeric',
    'Partial',
    'PyTree',
    'Real',
    'RealArray',
    'RealNumeric',
    'Shape',
    'ShapeLike',
    'SliceLike',
    'TapFunctionTransforms',
    'abs_square',
    'abstract_custom_jvp',
    'abstract_jit',
    'assert_tree_allclose',
    'copy_cotangent',
    'cotangent_combinator',
    'create_diagonal_array',
    'custom_jvp',
    'custom_jvp_method',
    'custom_vjp',
    'custom_vjp_method',
    'dataclasses',
    'default_atol',
    'default_rtol',
    'default_tols',
    'diffused_leaky_integrate',
    'display_generic',
    'divide_nonnegative',
    'divide_where',
    'get_relative_test_string',
    'get_test_string',
    'gradient',
    'graph_arrow',
    'graph_edge_name',
    'hessian',
    'internal_print_generic',
    'inverse_softplus',
    'jit',
    'leaky_covariance',
    'leaky_data_weight',
    'leaky_integrate',
    'leaky_integrate_time_series',
    'matrix_dot_product',
    'matrix_vector_mul',
    'outer_product',
    'print_cotangent',
    'print_generic',
    'register_graph_as_jax_pytree',
    'register_graph_as_nnx_node',
    'replace_cotangent',
    'scale_cotangent',
    'softplus',
    'tree_allclose',
    'zero_tangent_like',
]



from __future__ import annotations

from collections.abc import Sequence
from typing import Any

import numpy as np
import numpy.typing as npt
from jax import Array as JaxArray

Shape = tuple[int, ...]
ShapeLike = int | Sequence[int]
_SliceLikeItem = int | None | slice
SliceLike = _SliceLikeItem | tuple[_SliceLikeItem, ...]


JaxBooleanArray = JaxArray
JaxIntegralArray = JaxArray
JaxRealArray = JaxArray
JaxComplexArray = JaxArray
KeyArray = JaxArray


NumpyArray = npt.NDArray[Any]
NumpyBooleanArray = npt.NDArray[np.bool_]
NumpyIntegralArray = npt.NDArray[np.integer[Any]]
NumpyRealArray = npt.NDArray[np.floating[Any]]
NumpyComplexArray = npt.NDArray[np.floating[Any] | np.complexfloating[Any, Any]]
Array = NumpyArray | JaxArray
BooleanArray = NumpyBooleanArray | JaxArray
IntegralArray = NumpyIntegralArray | JaxArray
RealArray = NumpyRealArray | JaxArray
ComplexArray = NumpyComplexArray | JaxArray


Integral = int
Real = float | int
Complex = complex | float | int
NumpyBooleanNumeric = NumpyBooleanArray | bool
NumpyIntegralNumeric = NumpyIntegralArray | Integral
NumpyRealNumeric = NumpyRealArray | Real
NumpyComplexNumeric = NumpyComplexArray | Complex
BooleanNumeric = BooleanArray | bool
IntegralNumeric = IntegralArray | Integral
RealNumeric = RealArray | Real
ComplexNumeric = ComplexArray | Complex


PyTree = Any
TapFunctionTransforms = Sequence[tuple[str, dict[str, Any]]]



from __future__ import annotations

import math
from collections.abc import Generator, Mapping, MutableSet
from contextlib import contextmanager
from dataclasses import fields, is_dataclass
from functools import singledispatch
from numbers import Number
from types import FunctionType
from typing import Any

import numpy as np
from jax import Array
from jax.errors import TracerArrayConversionError
from jax.tree_util import PyTreeDef
from jaxlib.xla_extension import PjitFunction
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from tjax.dataclasses import DataclassInstance

from ..annotations import NumpyArray
from .colors import solarized

# Constants ----------------------------------------------------------------------------------------
# Numeric (warm)
_numpy_array_color = solarized['magenta']
_jax_array_color = solarized['orange']
_number_color = solarized['yellow']
# solarized['red']
# Classes (cool)
_class_color = solarized['violet']
# solarized['cyan']
_type_color = solarized['green']
_string_color = solarized['base0']
# Other
_key_color = solarized['blue']
_separator_color = solarized['base00']
_table_color = solarized['base02']
# _unknown_color = f"{solarized['red']} bold"
_seen_color = solarized['red']


# Extra imports ------------------------------------------------------------------------------------
FlaxModule: type[Any]
FlaxVariable: type[Any]
FlaxState: type[Any]
try:
    from flax import nnx
except ImportError:
    flax_loaded = False
    def is_node_type(x: type[Any]) -> bool:
        del x
        return False
    FlaxModule = type(None)
    FlaxVariable = type(None)
    FlaxState = type(None)
else:
    flax_loaded = True
    FlaxModule = nnx.Module
    FlaxVariable = nnx.Variable
    FlaxState = nnx.State
    is_node_type = nnx.graph.is_node_type


def attribute_filter(value: Any, attribute_name: str) -> bool:
    is_private = attribute_name.startswith('_')
    if flax_loaded:
        from flax import nnx  # noqa: PLC0415
        if isinstance(value, nnx.State) and is_private:
            return False
        if (isinstance(value, nnx.Variable | nnx.VariableState)
            and (is_private or attribute_name.endswith('_hooks'))):
            return False
        if isinstance(value, FlaxModule) and attribute_name.startswith('_graph_node__'):
            return False
    return True


@singledispatch
def display_generic(value: Any,
                    *,
                    seen: MutableSet[int] | None = None,
                    key: str = '',
                    ) -> Tree:
    if seen is None:
        seen = set()
    with _verify(value, seen, key) as x:
        if x:
            return x
        if is_dataclass(value) and not isinstance(value, type):
            return _display_dataclass(value, seen=seen, key=key)  # type: ignore[unreachable]
        return _display_object(value, seen=seen, key=key)
    # _assemble(key, Text(str(value), style=_unknown_color))


@display_generic.register
def _(value: str,
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    # No need for seen set ellision.
    return _assemble(key, Text(f'"{value}"', style=_string_color))


@display_generic.register(type)
def _(value: type[Any],
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    # No need for seen set ellision.
    return _assemble(key, Text(f"type[{value.__name__}]", style=_type_color))


@display_generic.register(np.ndarray)
def _(value: NumpyArray,
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    # No need for seen set ellision.
    retval = _assemble(key,
                       Text(f"NumPy Array {value.shape} {value.dtype}", style=_numpy_array_color))
    _show_array(retval, value)
    return retval


@display_generic.register
def _(value: Array,
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    # No need for seen set ellision.
    retval = _assemble(key,
                       Text(f"Jax Array {value.shape} {value.dtype}",
                            style=_jax_array_color))
    try:
        np_value = np.asarray(value)
    except TracerArrayConversionError:
        pass  # This happens when trying to print a tracer in immediate mode.
    else:
        _show_array(retval, np_value)
    return retval


@display_generic.register(type(None))
@display_generic.register(Number)
def _(value: None | Number,
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    # No need for seen set ellision.
    return _assemble(key, Text(str(value), style=_number_color))


@display_generic.register(Mapping)
def _(value: Mapping[Any, Any],
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    if seen is None:
        seen = set()
    with _verify(value, seen, key) as x:
        if x:
            return x
        retval = display_class(key, type(value))
        for sub_key, sub_value in value.items():
            retval.children.append(display_generic(sub_value, seen=seen, key=str(sub_key)))
        return retval


@display_generic.register(tuple)
@display_generic.register(list)
def _(value: tuple[Any, ...] | list[Any],
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    if seen is None:
        seen = set()
    with _verify(value, seen, key) as x:
        if x:
            return x
        retval = display_class(key, type(value))
        for sub_value in value:
            retval.children.append(display_generic(sub_value, seen=seen, key=""))
        return retval


@display_generic.register(PyTreeDef)
def _(value: PyTreeDef,
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    if seen is None:
        seen = set()
    with _verify(value, seen, key) as x:
        if x:
            return x
        retval = display_class(key, type(value))
        retval.children.append(display_generic(hash(value), seen=seen, key="hash"))
        return retval


@display_generic.register(FunctionType)
@display_generic.register(PjitFunction)
def _(value: PyTreeDef,
      *,
      seen: MutableSet[int] | None = None,
      key: str = '',
      ) -> Tree:
    if seen is None:
        seen = set()
    with _verify(value, seen, key) as x:
        if x:
            return x
        name = getattr(value, '__qualname__', "")
        retval = display_class(key, type(value))
        retval.children.append(display_generic(name, seen=seen, key="name"))
        return retval


# Public unexported functions ----------------------------------------------------------------------
def display_class(key: str, cls: type[Any]) -> Tree:
    name = cls.__name__
    tags = []
    if is_dataclass(cls):
        tags.append('dataclass')
    if flax_loaded and cls not in {int, float, tuple, list, set, dict}:
        if issubclass(cls, FlaxModule):
            tags.append('flax-module')
        elif is_node_type(cls):
            tags.append('flax-node')
    if tags:
        name += f"[{','.join(tags)}]"
    return _assemble(key, Text(name, style=_class_color))


# Private functions --------------------------------------------------------------------------------
def _display_dataclass(value: DataclassInstance,
                       *,
                       seen: MutableSet[int],
                       key: str = '',
                       ) -> Tree:
    retval = display_class(key, type(value))
    names = set()
    for field_info in fields(value):
        name = field_info.name
        if not field_info.repr:
            continue
        names.add(name)
        if not attribute_filter(value, name):
            continue
        display_name = name
        if not field_info.init:
            display_name += ' (module)'
        sub_value = getattr(value, name, None)
        retval.children.append(display_generic(sub_value, seen=seen, key=display_name))
    variables = _variables(value)
    for name, sub_value in variables.items():
        if name in names or not attribute_filter(value, name):
            continue
        retval.children.append(display_generic(sub_value, seen=seen, key=name + '*'))
    return retval


def _display_object(value: Any,
                    *,
                    seen: MutableSet[int],
                    key: str = '',
                    hide_private: bool = False,
                    ) -> Tree:
    retval = display_class(key, type(value))
    variables = _variables(value)
    for name, sub_value in variables.items():
        if not attribute_filter(value, name):
            continue
        retval.children.append(display_generic(sub_value, seen=seen, key=name))
    return retval


def _variables(value: Any) -> dict[str, Any]:
    try:
        variables = vars(value)
    except TypeError:
        variables = ({name: getattr(value, name) for name in value.__slots__}
                     if hasattr(value, '__slots__')
                     else {})
    return {key: value
            for key, value in variables.items()
            if key != '_module__state'
            if not key.startswith('__')}


@contextmanager
def _verify(value: Any,
            seen: MutableSet[int],
            key: str
            ) -> Generator[Tree | None, None, None]:
    if id(value) in seen:
        yield _assemble(key, Text(f'<seen {id(value)}>', style=_seen_color))
        return
    seen.add(id(value))
    yield None
    seen.remove(id(value))


def _assemble(key: str,
              type_text: Text,
              separator: str = '=') -> Tree:
    """Returns: A Rich Tree for a given key-value pair."""
    if key:
        return Tree(Text.assemble(Text(key, style=_key_color),
                                  Text(separator, style=_separator_color),
                                  type_text))
    return Tree(type_text)


@singledispatch
def _format_number(x: Any) -> str:
    return str(x)


@_format_number.register(np.inexact)
def _(x: np.inexact[Any]) -> str:
    return f"{x:.4f}"


def _show_array(tree: Tree, array: NumpyArray) -> None:
    """Add a representation of array to the Rich tree."""
    if not issubclass(array.dtype.type, np.bool_ | np.number):
        return
    if math.prod(array.shape) == 0:
        return
    if 1 in array.shape:
        _show_array(tree,
                    array[tuple[int | slice, ...](0 if s == 1 else slice(None)
                                                        for s in array.shape)])
        return
    if any(x > 12 for x in array.shape) or len(array.shape) > 2:  # noqa: PLR2004
        xarray = np.asarray(array)
        tree.children.append(display_generic(float(np.mean(xarray)), seen=set(), key="mean"))
        tree.children.append(display_generic(float(np.std(xarray)), seen=set(), key="deviation"))
        return
    if len(array.shape) == 0:
        tree.add(_format_number(array[()]))
        return
    table = Table(show_header=False,
                  show_edge=False,
                  style=_table_color)
    for _ in range(array.shape[-1]):
        table.add_column()
    if len(array.shape) == 1:
        table.add_row(*(_format_number(array[i])
                        for i in range(array.shape[0])))
    elif len(array.shape) == 2:  # noqa: PLR2004
        for j in range(array.shape[0]):
            table.add_row(*(_format_number(array[j, i])
                            for i in range(array.shape[1])))
    tree.add(table)



from __future__ import annotations

from typing import Any

from rich.console import Console
from rich.tree import Tree

from .display_generic import display_generic

# Global variables ---------------------------------------------------------------------------------
global_console = Console()


# Functions ----------------------------------------------------------------------------------------
def internal_print_generic(*args: Any,
                  raise_on_nan: bool = True,
                  console: Console | None = None,
                  **kwargs: Any) -> None:
    if console is None:
        console = global_console
    found_nan = False
    root = Tree("", hide_root=True)
    for value in args:
        root.add(display_generic(value, seen=set()))
        found_nan = found_nan or raise_on_nan and 'nan' in str(root)
    for key, value in kwargs.items():
        root.add(display_generic(value, seen=set(), key=key))
        found_nan = found_nan or raise_on_nan and 'nan' in str(root)
    console.print(root)
    if found_nan:
        assert False  # noqa: PT015



from __future__ import annotations

from typing import TypeVar, cast, overload

import numpy as np
from array_api_compat import get_namespace

from .annotations import (BooleanArray, ComplexArray, IntegralArray, JaxArray, JaxComplexArray,
                          JaxRealArray, RealArray)


@overload
def abs_square(x: JaxComplexArray) -> JaxRealArray:
    ...


@overload
def abs_square(x: ComplexArray) -> RealArray:
    ...


def abs_square(x: ComplexArray) -> RealArray:
    xp = get_namespace(x)
    # TODO: remove workaround when Jax is 0.4.27.
    return xp.square(x.real) + xp.square(xp.asarray(x.imag))


# TODO: Remove this when the Array API has it with broadcasting under xp.linalg.norm.
@overload
def outer_product(x: JaxRealArray, y: JaxRealArray) -> JaxRealArray:
    ...


@overload
def outer_product(x: RealArray, y: RealArray) -> RealArray:
    ...


def outer_product(x: RealArray, y: RealArray) -> RealArray:
    """Return the broadcasted outer product of a vector with itself.

    This is xp.einsum("...i,...j->...ij", x, y).
    """
    xp = get_namespace(x, y)
    xi = xp.reshape(x, (*x.shape, 1))
    yj = xp.reshape(y.conjugate(), (*y.shape[:-1], 1, y.shape[-1]))
    return xi * yj


@overload
def matrix_vector_mul(x: JaxRealArray, y: JaxRealArray) -> JaxRealArray:
    ...


@overload
def matrix_vector_mul(x: RealArray, y: RealArray) -> RealArray:
    ...


def matrix_vector_mul(x: RealArray, y: RealArray) -> RealArray:
    """Return the matrix-vector product.

    This is xp.einsum("...ij,...j->...i", x, y).

    Note the speed difference:
    * 14.3 µs: xp.vecdot(matrix_vector_mul(m, x), x)
    * 4.44 µs: np.einsum("...i,...ij,...j->...", x, m, x)
    """  # noqa: RUF002
    xp = get_namespace(x, y)
    y = xp.reshape(y, (*y.shape[:-1], 1, y.shape[-1]))
    return xp.sum(x * y, axis=-1)


@overload
def matrix_dot_product(x: JaxRealArray, y: JaxRealArray) -> JaxRealArray:
    ...


@overload
def matrix_dot_product(x: RealArray, y: RealArray) -> RealArray:
    ...


def matrix_dot_product(x: RealArray, y: RealArray) -> RealArray:
    """Return the "matrix dot product" of a matrix with the outer product of a vector.

    This equals:
    * 1.19 µs: xp.einsum("...ij,...ij", x, y)
    * 1.77 µs: xp.sum(x * y, axis=(-2, -1))
    # 3.87 µs: xp.linalg.trace(xp.moveaxis(x, -2, -1) @ y)
    """  # noqa: RUF002
    xp = get_namespace(x, y)
    return xp.sum(x * y, axis=(-2, -1))


def divide_where(dividend: ComplexArray,
                 divisor: ComplexArray | IntegralArray,
                 *,
                 where: BooleanArray | None = None,
                 otherwise: ComplexArray | None = None) -> ComplexArray:
    """Return the quotient or a special value when a condition is false.

    Returns: `xp.where(where, dividend / divisor, otherwise)`, but without evaluating
    `dividend / divisor` when `where` is false.  This prevents some exceptions.
    """
    if where is None:
        assert otherwise is None
        xp = get_namespace(dividend, divisor)
        return xp.divide(dividend, divisor)
    assert otherwise is not None
    xp = get_namespace(dividend, divisor, where, otherwise)
    dividend = xp.where(where, dividend, 1.0)
    divisor = xp.where(where, divisor, 1.0)
    quotient: ComplexArray = xp.divide(dividend, divisor)
    return xp.where(where, quotient, otherwise)


@overload
def divide_nonnegative(dividend: JaxRealArray, divisor: JaxRealArray) -> JaxRealArray:
    ...


@overload
def divide_nonnegative(dividend: RealArray, divisor: RealArray) -> RealArray:
    ...


def divide_nonnegative(dividend: RealArray, divisor: RealArray) -> RealArray:
    """Quotient for use with positive reals that never returns NaN.

    Returns: The quotient assuming that the dividend and divisor are nonnegative, and infinite
    whenever the divisor equals zero.
    """
    xp = get_namespace(dividend, divisor)
    return cast(RealArray, divide_where(dividend, divisor, where=divisor > 0.0,
                                        otherwise=xp.asarray(xp.inf)))


# Remove when https://github.com/scipy/scipy/pull/18605 is released.
@overload
def softplus(x: JaxRealArray) -> JaxRealArray:
    ...


@overload
def softplus(x: RealArray) -> RealArray:
    ...


def softplus(x: RealArray) -> RealArray:
    xp = get_namespace(x)
    return xp.logaddexp(xp.asarray(0.0), x)


@overload
def inverse_softplus(y: JaxRealArray) -> JaxRealArray:
    ...


@overload
def inverse_softplus(y: RealArray) -> RealArray:
    ...


def inverse_softplus(y: RealArray) -> RealArray:
    xp = get_namespace(y)
    return xp.where(y > 80.0,  # noqa: PLR2004
                    y,
                    xp.log(xp.expm1(y)))


T = TypeVar('T', bound=ComplexArray)


def create_diagonal_array(m: T) -> T:
    """A vectorized version of diagonal.

    Args:
        m: Has shape (*k, n)
    Returns: Array with shape (*k, n, n) and the elements of m on the diagonals.
    """
    xp = get_namespace(m)
    pre = m.shape[:-1]
    n = m.shape[-1]
    s = (*m.shape, n)
    retval = xp.zeros((*pre, n ** 2), dtype=m.dtype)
    for index in np.ndindex(*pre):
        target_index = (*index, slice(None, None, n + 1))
        source_values = m[*index, :]  # type: ignore[arg-type]
        if isinstance(retval, JaxArray):
            retval.at[target_index].set(source_values)
        else:
            retval[target_index] = source_values
    return xp.reshape(retval, s)



from __future__ import annotations

from collections.abc import Callable, Sequence
from functools import update_wrapper
from typing import Any, Concatenate, Generic, TypeVar, overload

import jax
from jax.tree_util import Partial
from typing_extensions import ParamSpec, Self, override

from .function_markers import all_wrapper_assignments

R_co = TypeVar('R_co', covariant=True)
F = TypeVar('F', bound=Callable[..., Any])
P = ParamSpec('P')
U = TypeVar("U")


def jit(func: F, **kwargs: Any) -> F:
    """A version of jax.jit that preserves flags.

    This ensures that abstract methods stay abstract, method overrides remain overrides.
    """
    retval = jax.jit(func, **kwargs)
    update_wrapper(retval, func, all_wrapper_assignments)
    # Return type is fixed by https://github.com/NeilGirdhar/jax/tree/jit_annotation.
    return retval  # type: ignore[return-value] # pyright: ignore


class custom_vjp(Generic[P, R_co]):  # noqa: N801
    """A shim class over jax.custom_vjp that uses ParamSpec.

    Args:
        func: The function to decorate.
        static_argnums: The indices of the **static** arguments -- nothing to do with
            differentiation.
    """

    vjp: jax.custom_vjp[R_co]

    @override
    def __init__(self,
                 func: Callable[P, R_co],
                 *,
                 static_argnums: tuple[int, ...] = ()):
        super().__init__()
        static_argnums = tuple(sorted(static_argnums))
        self.vjp = jax.custom_vjp(func, nondiff_argnums=static_argnums)
        update_wrapper(self, func, all_wrapper_assignments)

    def defvjp(self,
               fwd: Callable[P, tuple[R_co, Any]],
               bwd: Callable[..., Any]) -> None:
        self.vjp.defvjp(fwd, bwd)

    def __call__(self, /, *args: P.args, **kwargs: P.kwargs) -> R_co:
        return self.vjp(*args, **kwargs)


class custom_vjp_method(Generic[U, P, R_co]):  # noqa: N801
    """A shim class over jax.custom_vjp that uses ParamSpec and works with methods.

    Args:
        func: The method to decorate.
        static_argnums: The indices of the **static** arguments -- nothing to do with
            differentiation.
    """

    vjp: jax.custom_vjp[R_co]

    @override
    def __init__(self,
                 func: Callable[Concatenate[U, P], R_co],
                 *,
                 static_argnums: tuple[int, ...] = ()):
        super().__init__()
        static_argnums = tuple(sorted(static_argnums))
        self.vjp = jax.custom_vjp(func, nondiff_argnums=static_argnums)
        update_wrapper(self, func, all_wrapper_assignments)

    def defvjp(self,
               fwd: Callable[Concatenate[U, P], tuple[R_co, Any]],
               bwd: Callable[..., Any]) -> None:
        self.vjp.defvjp(fwd, bwd)

    def __call__(self,
                 u: U,
                 /,
                 *args: P.args,
                 **kwargs: P.kwargs) -> R_co:
        return self.vjp(u, *args, **kwargs)

    @overload
    def __get__(self, instance: None, owner: Any = None) -> Self:
        ...

    @overload
    def __get__(self, instance: U, owner: Any = None) -> Callable[P, R_co]:
        ...

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R_co]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)  # type: ignore[no-untyped-call]


class custom_jvp(Generic[P, R_co]):  # noqa: N801
    """A shim class over jax.custom_jvp that uses ParamSpec.

    Args:
        func: The function to decorate.
        nondiff_argnums: The indices of the non-differentiated arguments.
    """

    @override
    def __init__(self,
                 func: Callable[P, R_co],
                 *,
                 nondiff_argnums: tuple[int, ...] = ()):
        super().__init__()
        nondiff_argnums = tuple(sorted(nondiff_argnums))
        self.jvp = jax.custom_jvp(func, nondiff_argnums=nondiff_argnums)
        update_wrapper(self, func, all_wrapper_assignments)

    def defjvp(self, jvp: Callable[..., tuple[R_co, R_co]]) -> None:
        """Implement the custom forward pass of the custom derivative.

        Args:
            jvp: The custom forward pass.
        """
        self.jvp.defjvp(jvp)

    def __call__(self, /, *args: P.args, **kwargs: P.kwargs) -> R_co:
        return self.jvp(*args, **kwargs)


class custom_jvp_method(Generic[U, P, R_co]):  # noqa: N801
    """A shim class over jax.custom_jvp that uses ParamSpec and works with methods.

    Args:
        func: The method to decorate.
        nondiff_argnums: The indices of the non-differentiated arguments.
    """

    @override
    def __init__(self,
                 func: Callable[Concatenate[U, P], R_co],
                 *,
                 nondiff_argnums: tuple[int, ...] = ()):
        super().__init__()
        nondiff_argnums = tuple(sorted(nondiff_argnums))
        self.jvp = jax.custom_jvp(func, nondiff_argnums=nondiff_argnums)
        update_wrapper(self, func, all_wrapper_assignments)

    def defjvp(self, jvp: Callable[Concatenate[U, P], tuple[R_co, R_co]]) -> None:
        """Implement the custom forward pass of the custom derivative.

        Args:
            jvp: The custom forward pass.
        """
        self.jvp.defjvp(jvp)

    def __call__(self, u: U, /, *args: P.args, **kwargs: P.kwargs) -> R_co:
        return self.jvp(u, *args, **kwargs)

    @overload
    def __get__(self, instance: None, owner: Any = None) -> Self:
        ...

    @overload
    def __get__(self, instance: U, owner: Any = None) -> Callable[P, R_co]:
        ...

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R_co]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)  # type: ignore[no-untyped-call]


def hessian(fun: Callable[..., U],
            argnums: int | Sequence[int] = 0,
            *,
            has_aux: bool = False,
            holomorphic: bool = False,
            reverse_only: bool = False
            ) -> Callable[..., U]:
    if not reverse_only:
        return jax.hessian(fun, argnums=argnums, has_aux=has_aux, holomorphic=holomorphic)
    return jax.jacrev(jax.jacrev(fun, argnums, has_aux=has_aux, holomorphic=holomorphic),
                      argnums, has_aux=has_aux, holomorphic=holomorphic)
